import pandas as pd
from cachetools import TTLCache, cached
from komoutils.core import KomoBase

from aporacle import MONGODB_URI, MONGODB_NAME
from aporacle.data.db.crud import Crud


class SymbolData(KomoBase):
    def __init__(self):
        self.crud: Crud = Crud(uri=MONGODB_URI, db_name=MONGODB_NAME)

    @cached(cache=TTLCache(maxsize=1024, ttl=3600))
    def get_symbol(self, symbol: str) -> list:
        filters = {"asset": symbol}
        return self.crud.read_symbol_data(filters)

    def get(self, symbol: str) -> pd.DataFrame:
        df = pd.DataFrame(self.get_symbol(symbol))
        return df



import os

import requests


def get_api_setup_data():
    try:
        with requests.Session() as session:
            response = session.get(url=f"http://{os.getenv('TSO_SETUP_URL')}/setup")
            response.raise_for_status()
            setup = response.json()
            return setup

    except Exception as e:
        raise


setup_data = get_api_setup_data()

# Data Service Streamers and APIs
WS = 'ws://'
WSS = 'wss://'
HTTP = 'http://'
HTTPS = 'https://'

tso_utilities_streamer_endpoint = setup_data["algorithms"]["streamers"]["utilities"]["url"]
tso_prepared_streamer_endpoint = setup_data["algorithms"]["streamers"]["prepared"]["url"]
tso_trades_streamer_endpoint = setup_data["algorithms"]["streamers"]["trades"]["url"]

streams = {
    'utilities': tso_utilities_streamer_endpoint,
    'prepared': tso_prepared_streamer_endpoint,
    'trades': tso_trades_streamer_endpoint
}

tso_utility_topics = [topic for topic in setup_data["algorithms"]["tso"]["utility_topics"]]
print(tso_utility_topics)

# CHAIN VOTING ROUNDS
algorithm_prepared_data_buffer = 17
ml_predictions_buffer = 12
submission_commit_buffer = 7

# Web3 Providers
for chain in ['coston', 'songbird', 'flare']:
    try:
        web3_provider_list = setup_data["chains"][chain]["providers"]["rpc"]
        web3_websocket_list = setup_data["chains"][chain]["providers"]["ws"]
    except Exception as e:
        print(e)

test_web3_provider_list = [os.getenv("WEB3_PROVIDER")]
test_web3_websocket_list = [os.getenv("WEB3_WEBSOCKETS")]

# Chain Wallet Configuration
flare_reward_offers_manager_addresses = {}
for chain in ['coston', 'songbird', 'flare']:
    try:
        flare_reward_offers_manager_addresses[chain] \
            = setup_data["chains"][chain]["flare_reward_offers_manager_address"]
    except Exception as e:
        print(e)



import logging
import pickle
import uuid
from typing import Optional, List

import numpy as np
import pandas as pd
import pendulum
from komoutils.core import KomoBase
from komoutils.core.time import the_time_in_iso_now_is
from pydantic import BaseModel
from sklearn.linear_model import Ridge
from sklearn.model_selection import train_test_split


def get_rmse(y_test, y_pred):
    rmse = np.sqrt(np.mean((y_test - y_pred) ** 2))
    # print(y_test-y_pred)
    return rmse


def get_r2(y_test, y_pred):
    r2 = 1 - (np.sum((y_test - y_pred) ** 2) / np.sum((y_test - y_test.mean()) ** 2))
    return r2


class ModelResult(BaseModel):
    executor_id: str
    chain: str
    tso: str
    label: str
    target: list
    rmse: float
    r2: float
    model: bytes
    model_version: str
    model_type: str
    parameters: dict
    features: list
    symbols: list
    data_size: int
    time: str


class Runner(KomoBase):
    def __init__(self, tso: str, target: str):
        self.tso: Optional[str] = tso
        self.target: Optional[str] = target
        self.X = None
        self.y = None
        self.X_train = self.X_test = self.y_train = self.y_test = None
        self.train_size = 0.8
        self.symbols: Optional[list] = None

        self.results: list = []

    def _preprocess_inputs(self):
        try:
            # Train-test split
            X_train, X_test, y_train, y_test = train_test_split(self.X,
                                                                self.y,
                                                                train_size=self.train_size,
                                                                shuffle=True,
                                                                random_state=1)

            return X_train, X_test, y_train, y_test
        except Exception as e:
            self.log_with_clock(log_level=logging.ERROR, msg=f"Error during input processing. {e}")
            return None, None, None, None

    def _train(self, alpha: float = 0.1):
        try:
            name = 'ridge'
            parameters = {
                'alpha': alpha,
            }
            model = Ridge(alpha=parameters['alpha'])

            model.fit(self.X_train, self.y_train)
            y_pred = model.predict(self.X_test)
            y_test = self.y_test.to_numpy().reshape(1, -1)

            r2 = get_r2(y_test, y_pred)
            rmse = get_rmse(y_test, y_pred)

            chain = "combined"
            model_file_name = f"model_{self.key}_{self.target}_{name}".lower()

            pickled_model = pickle.dumps(model)
            result = {
                "executor_id": uuid.uuid4().hex,
                "chain": chain,
                "asset": self.tso,
                "label": model_file_name,
                "target": self.target,
                "rmse": rmse,
                "r2": r2,
                "model": pickled_model,
                "model_version": f"{chain}_{pendulum.now().timestamp()}",
                "model_type": name,
                "parameters": parameters,
                "features": self.X.columns.to_list(),
                "symbols": sorted(self.symbols),
                "data_size": len(self.X),
                "time": the_time_in_iso_now_is()
            }

            self.results.append(ModelResult(**result))

        except Exception as e:
            self.log_with_clock(log_level=logging.ERROR, msg=f"Error during training. {e}")
            return None

    def run(self, input_data: pd.DataFrame, symbols: list, alphas: list) -> List[ModelResult]:
        # unpack columns looking for 'features' and 'target' tags.
        columns_names = input_data.columns.to_list()
        features = [col for col in columns_names if 'features' in col.lower()]
        target = [col for col in columns_names if 'target' in col.lower()]

        self.X = input_data[features]
        self.y = input_data[target]

        self.symbols = symbols

        for alpha in alphas:
            self._train(alpha=alpha)

        return self.results



from komoutils.db.mongodb_reader_writer import MongoDBReaderWriter


class Crud(MongoDBReaderWriter):
    def __init__(self, uri: str, db_name: str):
        super().__init__(uri, db_name)

    def read_symbol_data(self, filters=None):
        if filters is None:
            filters = {}
        return self.read(filters)



from typing import List, Optional

import pandas as pd
from komoutils.core import KomoBase

from aporacle.data.asset import AssetData
from aporacle.data.symbols import SymbolData


class Combination(KomoBase):
    def __init__(self, symbols: List[str] = None):
        self.symbols: Optional[list] = symbols
        self.combined: Optional[pd.DataFrame] = None
        self.symbol_data: SymbolData = SymbolData()
        self.asset_data: AssetData = AssetData()

    def combine(self, data: List[pd.DataFrame]):
        self.combined = pd.concat(data, axis=1)
        return self.combined

    def get(self):
        data: List[pd.DataFrame] = [self.symbol_data.get(symbol) for symbol in self.symbols]
        return self.combine(data)



import logging
import os
from os.path import basename, dirname, abspath
from typing import Optional

from komoutils.core.time import the_time_in_iso_now_is
from komoutils.logger import KomoLogger

class SharedSetup:
    _shared_instance: "SharedSetup" = None
    _logger: Optional[KomoLogger] = None

    @classmethod
    def get_instance(cls) -> "SharedSetup":
        if cls._shared_instance is None:
            cls._shared_instance = SharedSetup()
        return cls._shared_instance

    @classmethod
    def logger(cls) -> KomoLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def log_with_clock(self, log_level: int, msg: str, **kwargs):
        self.logger().log(log_level, f"{self.__class__.__name__} {msg} [clock={str(the_time_in_iso_now_is())}]",
                          **kwargs)

    def __init__(self):
        self.algorithm_name: str = ""
        self.environment: str = os.getenv("ENVIRONMENT")
        self.trades_streamer_endpoint: str = f'ws://{os.getenv("TRADES_STREAMER_ENDPOINT")}'
        print(self.trades_streamer_endpoint)

    def start(self):
        self.log_with_clock(log_level=logging.INFO, msg=f"Shared algorithm setup service started. ")

        try:
            # Get the name of the algorithm for dir name
            self.algorithm_name = str(basename(dirname(abspath(__file__)))).replace("_", "-")

        except Exception as e:
            raise e

        self.log_with_clock(log_level=logging.INFO, msg=f"Shared algorithm setup service successfully started. ")



import json
import logging
from typing import Optional

import websockets
from komoutils.core import KomoBase

from aporacle.conf import WS, streams


class DataBroadcaster(KomoBase):
    def __init__(self, stream: str = 'utilities'):
        self.url: str = f"{WS}{streams[stream]}/ingest"

    @property
    def name(self):
        return "data_broadcaster"

    async def send(self, data: list):
        try:
            assert type(data) == list, "Data must be a list of dictionaries."
            async with websockets.connect(self.url) as ws:
                for d in data:
                    await ws.send(json.dumps(d))
        except AssertionError as ae:
            self.log_with_clock(log_level=logging.ERROR, msg=f"Data format error. {ae}")
        except Exception as e:
            raise e

